From 5d49eebfdf4af6bd8ebd3ba79e43c47476b1c15f Mon Sep 17 00:00:00 2001
From: Sebastian Pop <spop@amazon.com>
Date: Thu, 23 Dec 2021 22:30:30 +0000
Subject: [PATCH] [arm64] port scanPosLast

           scanPosLast	5.08x 	 842.11   	 4277.83
---
 source/common/aarch64/asm-primitives.cpp |   1 +
 source/common/aarch64/fun-decls.h        |   1 +
 source/common/aarch64/pixel-util.S       | 122 +++++++++++++++++++++++
 3 files changed, 124 insertions(+)

diff --git a/source/common/aarch64/asm-primitives.cpp b/source/common/aarch64/asm-primitives.cpp
index 868535e3e..1d6b5e524 100644
--- a/source/common/aarch64/asm-primitives.cpp
+++ b/source/common/aarch64/asm-primitives.cpp
@@ -664,6 +664,7 @@ void setupAssemblyPrimitives(EncoderPrimitives &p, int cpuMask)
 
         p.weight_pp = PFX(weight_pp_neon);
         p.weight_sp = PFX(weight_sp_neon);
+        p.scanPosLast = PFX(scanPosLast_neon);
 #endif
 
         // quant
diff --git a/source/common/aarch64/fun-decls.h b/source/common/aarch64/fun-decls.h
index 3df19b59c..1925728c4 100644
--- a/source/common/aarch64/fun-decls.h
+++ b/source/common/aarch64/fun-decls.h
@@ -220,3 +220,4 @@ int PFX(psyCost_4x4_neon)(const pixel* source, intptr_t sstride, const pixel* re
 int PFX(psyCost_8x8_neon)(const pixel* source, intptr_t sstride, const pixel* recon, intptr_t rstride);
 void PFX(weight_pp_neon)(const pixel* src, pixel* dst, intptr_t stride, int width, int height, int w0, int round, int shift, int offset);
 void PFX(weight_sp_neon)(const int16_t* src, pixel* dst, intptr_t srcStride, intptr_t dstStride, int width, int height, int w0, int round, int shift, int offset);
+int PFX(scanPosLast_neon(const uint16_t *scan, const coeff_t *coeff, uint16_t *coeffSign, uint16_t *coeffFlag, uint8_t *coeffNum, int numSig, const uint16_t* scanCG4x4, const int trSize));
diff --git a/source/common/aarch64/pixel-util.S b/source/common/aarch64/pixel-util.S
index 1859053d2..c6bbb5c23 100644
--- a/source/common/aarch64/pixel-util.S
+++ b/source/common/aarch64/pixel-util.S
@@ -2402,3 +2402,125 @@ function PFX(weight_sp_neon)
     cbnz            x5, .loopHspUS
     ret
 endfunc
+
+// int scanPosLast(const uint16_t *scan x0, const coeff_t *coeff x1, uint16_t *coeffSign x2, uint16_t *coeffFlag x3, uint8_t *coeffNum x4, int numSig x5, const uint16_t* scanCG4x4 x6, const int trSize x7)
[MC] Suggest put every parameter in single line, such as
    const uint16_t *scan,   // x0
    const coeff_t *coeff,   // x1

+function PFX(scanPosLast_neon)
+    // convert unit of Stride(trSize) to int16_t
+    add             x7, x7, x7
+
+    // load scan table and convert to Byte
+    // v0 - Zigzag scan table
+    ldp             q0, q1, [x6]
+    xtn             v0.8b, v0.8h
+    xtn2            v0.16b, v1.8h
+
+    movrel          x10, g_SPL_and_mask
+    ldr             q28, [x10]
+
+    // clear CG count
+    mov             x9, #0
+
+.loop_spl:
+    // position of current CG
+    ldrh            w6, [x0]
[MC] Why not 'ldrh w6,[x0],#32' ?

+    add             x6, x6, x6
+    add             x6, x6, x1
[MC] Why not 'add x6,x1,x6,lsl #1' ?

+    add             x0, x0, #32
+
+    // loading current CG
+    ldr             d2, [x6]
+    ldr             d3, [x6, x7]
+    add             x14, x7, x7     // 2*x7
[MC] x14 is constant in the loop

+    ldr             d4, [x6, x14]
+    add             x15, x14, x7    // 3*x7
[MC] Same issue as x14

+    ldr             d5, [x6, x15]
+    ins             v2.d[1], v3.d[0]
+    ins             v4.d[1], v5.d[0]
[MC] INS is alias MOV

+    sqxtn           v2.8b, v2.8h
+    sqxtn2          v2.16b, v4.8h
+
+    // Zigzag
+    tbl             v3.16b, {v2.16b}, v0.16b
+
+    // get sign
+    cmeq            v5.16b, v3.16b, #0  //      equal to zero
+    mvn             v5.16b, v5.16b      // v5 = non-zero
[MC] Why not replace cmeq+mvn by cmgt?

+    cmlt            v3.16b, v3.16b, #0  // v3 = negative
+
+    // val - w13 = pmovmskb(v3)
+    and             v3.16b, v3.16b, v28.16b
+    mov             d4, v3.d[1]
+    addv            b13, v3.8b
+    addv            b14, v4.8b
[MC] ADDV support .16b?

+    fmov            w13, s13
+    fmov            w14, s14
+    orr             w13, w13, w14, lsl #8
+
+    // mask - w15 = pmovmskb(v5)
+    and             v5.16b, v5.16b, v28.16b
+    mov             d6, v5.d[1]
+    addv            b15, v5.8b
+    addv            b16, v6.8b
+    fmov            w15, s15
+    fmov            w14, s16
+    orr             w15, w15, w14, lsl #8
+
+    // coeffFlag = reverse_bit(w15) in 16-bit
+    rbit            w6, w15
+    lsr             w6, w6, #16
+    fmov            s30, w6
[MC] may reorder since we don't care elements order inside

+    add             x10, x9, x9
+    strh            w6, [x3, x10]
[MC] How about [x3,x9,lsl #1]?

+
+    // w14 = pext(val: w13, mask: w15)
+    cbz             w15, .pext_end
+    mov             x14, xzr
+    mov             w6, #1
+
+.pext:
+    and             w10, w15, w13
+    neg             w11, w15
+    tst             w10, w11
+    sub             w12, w15, #1
+    csel            w10, wzr, w6, eq
+    ands            w15, w12, w15
+    orr             w14, w10, w14
+    lsl             w6, w6, #1
+    b.ne            .pext
+.pext_end:

[MC] We try to collection bits depends on bitMasks, so we are unnecessary so complicated simulate code for x86 PEXT

Reference

    // w14 = pext(val: w13, mask: w15)
    // Used Regs x12-x15
    // (w13 & w15 will BROKEN after the loop)
    mov     w14, wzr;
.loop:
    cbz     w15, .pext_end;   // may accelerate by prepare with (w13 & w15)
    clz     w12, w15
    lsl     w13, w13, w12
    lsl     w15, w15, w12
    extr    w14, w14, w13, #31
    bfc     w15, #31, #1
    b      .loop

.pext_end:


+
+    add             x10, x9, x9
+    strh            w14, [x2, x10]
[MC] Same as above

+
+    // compute coeffNum = popcount(coeffFlag)
+    cnt             v30.8b, v30.8b
+    addv            b30, v30.8b
[MC] there only 2 bytes we care,
     addp throughput 3/2, adv throughtput 7/1

+    fmov            w6, s30
+    strb            w6, [x4, x9]
[MC] may reduce after above optimize

+    add             x9, x9, #1
+
+    sub             x5, x5, x6
+    cmp             x5, #0
+    bgt             .loop_spl
[MC] CBZ?

+
+    // fixup last CG non-zero flag
+    sub             x9, x9, #1
+    add             x10, x9, x9
+    ldrh            w11, [x3, x10]
[MC] if we allocate register careful above, we unnecessary load the result again

+    // count trailing zeros
+    rbit            w12, w11
+    clz             w12, w12
+    lsr             w11, w11, w12
+    strh            w11, [x3, x10]
+
+    // get last pos
+    lsl             x0, x9, #4
[MC] may merge into latest ADD

+    eor             w12, w12, #15
+    add             x0, x0, x12
+    ret
+endfunc
+
+
+const g_SPL_and_mask, align=8
+.byte 0x1, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x80, 0x1, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x80
+endconst
-- 
2.25.1

